"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
from cdecimal import Decimal
from buttonwood.MarketObjects.Side import Side
from buttonwood.MarketObjects.Side import BID_SIDE


class InvalidPriceException(Exception):
    pass


class Price:

    def __init__(self, price_value):
        """
        Represents a price. Is created with a string, for accuracy, or as a Decimal and is kept as a Decimal
  
        :param price_value: str
        """
        self._value = price_value if isinstance(price_value, Decimal) else Decimal(price_value)
        self.__hash = self._value.__hash__()

    def better_than(self, other_price, side):
        """
        Determines if this price is better than the passed in price. Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then better than means a higher price
           *if side is ask then better than means a lower price
           *if side is other then there is no such thing as a better price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        # if this price is a bid it is better than otherPrice if greater than

        if side.is_bid():
            return self > other_price
        # otherwise price is an ask and it is better than otherPrice if less than
        else:
            return self < other_price

    def better_or_same_as(self, other_price, side):
        """
        Determines if this price is better than or same as the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then better than means a higher price
           *if side is ask then better than means a lower price
           *if side is other then there is no such thing as a better price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        return True if other_price == self else self.better_than(other_price, side)

    def worse_than(self, other_price, side):
        """
        Determines if this price is worse than the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then worse than means a lower price
           *if side is ask then worse than means a higher price
           *if side is other then there is no such thing as a worse price so the return is False
        
        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        assert isinstance(other_price, Price), "other_price should be MarketObjects.Price.Price"
        assert isinstance(side, Side), "side should be MarketObjects.Side.Side"
        # if this price is a bid it is worse than otherPrice if less than
        if side.is_bid():
            return self < other_price
        # otherwise price is an ask and it is worse than otherPrice if greater than
        else:
            return self > other_price

    def worse_or_same_as(self, other_price, side):
        """
        Determines if this price is worse than or same as the passed in price.
        Returns True if it is and False if it isn't.
        The determination is made using the passed in side:
           *if side is bid then worse than means a lower price
           *if side is ask then worse than means a higher price
           *if side is other then there is no such thing as a worse price so the return is False

        :param other_price: Price. the price you are comparing to
        :param side: Side. the side used for the comparison
        :return: boolean
        """
        return True if other_price == self else self.worse_than(other_price, side)

    def ticks_behind(self, other_price, side, market):
        """
        For bids it subtracts its own price from the other price and divides by mpi
  
        For asks it subtracts the other price from its own price and divides by mpi.
  
        The result is if the other price is a better price then we get a positive
         number and if the other price is a worse price, we get a negative number.
  
        If otherprice is none it returns None
        :param other_price: MarketObjects.Price.Price the price you are comparing too
        :param side: MMarketObjects.Side.Side side of the market the comparison is happening on
        :param market: MarketObjects.Market.Market
        :return: float Can be None
        """
        if other_price is None:
            return None

        if side.is_bid():
            return (other_price - self) / market.mpi()
        else:
            return (self - other_price) / market.mpi()

    def __hash__(self):
        return self.__hash

    def __repr__(self):
        return self._value.__repr__()

    def __str__(self):
        return self._value.__str__()

    def __lt__(self, other):
        if not isinstance(other, Price):
            raise TypeError(other, '< requires another Price object')
        else:
            return self._value < other._value

    def __le__(self, other):
        if not isinstance(other, Price):
            raise TypeError(other, '<= requires another Price object')
        else:
            return self._value <= other._value

    def __eq__(self, other):
        if isinstance(other, Price):
            return self._value == other._value
        return False

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        if not isinstance(other, Price):
            raise TypeError(other, '> requires another Price object')
        else:
            return self._value > other._value

    def __ge__(self, other):
        if not isinstance(other, Price):
            raise TypeError(other, '>= requires another Price object')
        else:
            return self._value >= other._value

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self._value)

    def __add__(self, other):
        if isinstance(other, Price):
            other = other._value
        value = self._value + other
        return self.__class__(value)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Price):
            other = other._value
        value = self._value - other
        return self.__class__(value)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __mul__(self, other):
        if isinstance(other, Price):
            raise TypeError("Two prices cannot be multiplied")  # just doesn't make sense. Why would someone do this?
        value = self._value * other
        return self.__class__(value)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        return self.__truediv__(other)

    def __truediv__(self, other):
        if isinstance(other, Price):
            # not really sure why someone would want to divide a price by another price
            raise TypeError("Two prices cannot be divided")
        else:
            if other == 0:
                raise ZeroDivisionError()
            return self._value/ other

    def __floordiv__(self, other):
        if isinstance(other, Price):
            # not really sure why someone would want to divide a price by another price
            raise TypeError("Two prices cannot be divided")
        else:
            if other == 0:
                raise ZeroDivisionError()
            value = self._value // other
            return self.__class__(value)

    def __mod__(self, other):
        if isinstance(other, Price):
            # not really sure why someone would want to modulo a price by another price
            raise TypeError("modulo of two prices is not supported")
        if other == 0:
            raise ZeroDivisionError()
        return self._value % other

    def __divmod__(self, other):
        raise TypeError("divmod of a price is not supported")

    def __pow__(self, other):
        raise TypeError("pow of a price is not supported")

    def __neg__(self):
        return self.__class__(-self._value)

    def __pos__(self):
        return self.__class__(+self._value)

    def __abs__(self):
        return self.__class__(abs(self._value))

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return float(self._value)


class PriceFactory:

    def __init__(self, min_price_increment, min_price_increment_value=1, min_price=Price(-999999),
                 max_price=Price(999999)):
        assert isinstance(min_price_increment, (Decimal, str, int))
        assert isinstance(min_price_increment_value, (Decimal, str, int))
        self._mpi = Decimal(min_price_increment)
        self._mpi_value = Decimal(min_price_increment_value)
        self._min_price = min_price if isinstance(min_price, Price) else Price(min_price)
        self._max_price = max_price if isinstance(max_price, Price) else Price(max_price)
        assert self._min_price <= self._max_price, "Max price (%s) must be greater than or equal to min price (%s)." % \
                                                   (str(max_price), str(min_price))

    def next_price(self, price, side):
        return self.get_price(price + (self._mpi * (1 if side is BID_SIDE else -1)))

    def prev_price(self, price, side):
        return self.get_price(price + (self._mpi * (-1 if side is BID_SIDE else 1)))

    def get_price(self, price_value):
        price = price_value if isinstance(price_value, Price) else Price(price_value)
        if not self.is_valid_price(price):
            raise InvalidPriceException("%s is not valid price with min %s max %s and mpi %s" %
                                        (str(price_value), str(self._min_price), str(self._max_price), str(self._mpi)))
        return price

    def min_price_increment(self):
        """
        Gets the minimum price increment of a product. This is the smallest increment that the product trades in.
        :return: Decimal
        """
        return self._mpi

    def min_price_increment_value(self):
        """
        Gets the monetary value associated with the minimum price increment of a product.
        :return: Decimal
        """
        return self._mpi_value

    def mpi(self):
        """
        This is the same as `min_price_increment()'. It is just for the convenience of the developer.

        :return: Decimal
        """
        return self._mpi

    def mpi_value(self):
        """
        This is the same as `min_price_increment_value()'. It is just for the convenience of the developer.

        :return: Decimal
        """
        return self._mpi_value

    def is_valid_price(self, price):
        if not isinstance(price, Price):
            price = Price(price)
        return (self._min_price <= price <= self._max_price) and (price % self._mpi) == 0

    def to_json(self):
        return {"mpi": str(self._mpi), "mpv": str(self._mpi_value), "min": str(self._min_price),
                "max": str(self._max_price)}

    def __str__(self):
        return json.dumps(self.to_json())




"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.Price import Price
from buttonwood.MarketObjects.Side import Side


class PriceLevel(object):
    def __init__(self, price, visible_qty, hidden_qty=0, num_orders=None):
        """
        A price level is one price of one side of a book. It summarizes all the basic data found at that price, including
         the price itself, the visible qty, the hidden qty, and the number of orders. Hidden Qty is optional as many
         markets don't have it as an option. And number of orders is optional because based on the market that's data is
         being processed you might not have knowledge of number of orders.
  
         Visible Qty must be greater than 0 because can't have negative qty and if 0 then shouldn't be a price level.
  
         Hidden Qty must be >=0 because can't have a negative qty at a price level.
  
         Number of orders must be >= visible qty + hidden_qty because smallest size per order is 1 so can't have more
          orders than qty.
  
         If number of orders is defined then it has to be > 0 as can't be negative and 0 would mean no price level.
  
        :param price: MarketObjects.Price. The level's price
        :param visible_qty: int. quantity that is at the price and visible
        :param hidden_qty: int. quantity that is at the price and hidden. Optional. Defaults to 0.
        :param num_orders: int. the number of orders that makes up the price level. Optional. Defaults to None.
        """
        assert isinstance(price, Price)
        assert isinstance(visible_qty, int)
        assert isinstance(hidden_qty, int)
        assert visible_qty > 0
        assert hidden_qty >= 0
        assert num_orders is None or isinstance(num_orders, int)
        assert num_orders is None or num_orders <= visible_qty + hidden_qty
        assert num_orders is None or num_orders > 0
        self._price = price
        self._visible_qty = visible_qty
        self._hidden_qty = hidden_qty
        self._num_orders = num_orders

    def price(self):
        """
        The price of the price level
  
        :return: Price
        """
        return self._price

    def total_qty(self):
        """
        The total qty on the price level, which is visible + hidden
  
        :return: int.
        """
        return self._visible_qty + self._hidden_qty

    def visible_qty(self):
        """
        The visible qty at the price level
  
        :return: int
        """
        return self._visible_qty

    def hidden_qty(self):
        """
        The hidden qty at the price level
  
        :return: int
        """
        return self._hidden_qty

    def number_of_orders(self):
        """
        The number of orders that make up the price level.
  
        :return: int. Can be None.
        """
        return self._num_orders

    def better_than(self, other_level, side):
        """
        returns if this price level is better than another price level for a given side of the market
        :param side: Side
        :param other_level: PriceLevel
        :return: bool
        """
        assert isinstance(other_level, PriceLevel)
        assert isinstance(side, Side)
        return self.price().better_than(other_level.price(), side)

    def better_or_same_as(self, other_level, side):
        """
        returns if this price level is better than or same as another price level for a given side of the market
        :param side: Side
        :param other_level: PriceLevel
        :return: bool
        """
        assert isinstance(other_level, PriceLevel)
        assert isinstance(side, Side)
        return self.price().better_or_same_as(other_level.price(), side)

    def worse_than(self, other_level, side):
        """
        returns if this price level is worse than another price level for a given side of the market
        :param side: Side
        :param other_level: PriceLevel
        :return: bool
        """
        assert isinstance(other_level, PriceLevel)
        assert isinstance(side, Side)
        return self.price().worse_than(other_level.price(), side)

    def worse_or_same_as(self, other_level, side):
        """
        returns if this price level is worse than or same as another price level for a given side of the market
        :param side: Side
        :param other_level: PriceLevel
        :return: bool
        """
        assert isinstance(other_level, PriceLevel)
        assert isinstance(side, Side)
        return self.price().worse_or_same_as(other_level.price(), side)

    def __eq__(self, other):
        assert isinstance(other, PriceLevel)
        if other is None:
            return False
        if self.price() != other.price():
            return False
        if self.visible_qty() != other.visible_qty():
            return False
        if self.hidden_qty() != other.hidden_qty():
            return False
        if self.number_of_orders() is not None and other.number_of_orders() is not None:
            if self.number_of_orders() != other.number_of_orders():
                return False
        else:
            if self.number_of_orders() is None and other.number_of_orders() is not None:
                return False
            if self.number_of_orders() is not None and other.number_of_orders() is None:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        s = "%s : %d | %d" % (str(self._price), self.visible_qty(), self.hidden_qty())
        if self._num_orders is not None:
            s = s + " (%d)" % self._num_orders
        return s



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import defaultdict

"""
this isn't a true dict, but it wraps a dict and allows for easily initializing
 and using (in simple cases), an n deep dict, where the structure is:
  key1 -> key2 ->.... -> keyn -> payload
"""


def _establish_level(levels, default_value):
    if levels < 2:
        return defaultdict(default_value)
    return defaultdict(lambda: _establish_level(levels - 1, default_value))


class NDeepDict:
    def __init__(self, depth, default_value=None):
        assert depth > 0
        if default_value is None:
            default_value = lambda: None
        assert callable(default_value), "default_value must be not defined or callable"
        self._depth = depth
        self._dict = _establish_level(depth, default_value)

    def get(self, keys):
        """
        Pulling data out of the dict. Must have a number of args that is less
         than or equal to the depth of the dict.

        This returns what the args map to. If the same number of args as the
         defined depth it returns whatever object is in the final payload.
         Otherwise it returns the dict that the given keys lead to.

        :param keys: a list of the keys you want to use for accessing the dict.
        :return: the value of the passed in args
        """
        assert len(keys) <= self._depth, "To access a value out of dict hierarchy, number of keys must be less than or equal to depth of dict"
        result = self._dict
        for key in keys:
            result = result[key]
        return result

    def set(self, keys, value=None):
        """
        Setting data in the dict. This requires all keys that lead to final
         payload, so the number of args must equal the defined depth.

        Value defaults to None and if None then nothing happens.

        :param keys: a list of the keys you want to use for accessing the dict
        :param value: the value you want to set payload to. If None (the default), then nothing happens
        """
        assert len(keys) == self._depth, "Number of arguments must equal depth of dict to set a value"
        result = self._dict
        for key in keys[:-1]:
            result = result[key]
        result[keys[-1]] = value

    def inc(self, keys, amount=1):
        """
        Increment the payload of the dict. Note, if the payload isn't a numeric
         then this fails. This requires all keys that lead to final
         payload, so the number of args must equal the defined depth.

        amount is the amount to increment. It defaults to 1.

        :param keys: the keys you want to use for accessing the dict
        :param amount: amount to increment by, it defaults to 1.
        """
        assert len(keys) == self._depth, "Number of arguments must equal depth of dict to increment the payload value"
        result = self._dict
        for key in keys[:-1]:
            result = result[key]
        result[keys[-1]] += amount

    def dec(self, keys, amount=1):
        """
        Decrement the payload of the dict. Note, if the payload isn't a numeric
         then this fails. This requires all keys that lead to final
         payload, so the number of args must equal the defined depth.

        amount is the amount to decrement. It defaults to 1.

        :param keys: the keys you want to use for accessing the dict
        :param amount: amount to increment by, it defaults to 1.
        """
        assert len(keys) == self._depth, "Number of arguments must equal depth of dict to decrement the payload value"
        result = self._dict
        for key in keys[:-1]:
            result = result[key]
        result[keys[-1]] -= amount

    def _sum(self, v):
        total = 0
        if isinstance(v, dict):
            for value in v.itervalues():
                total += self._sum(value)
        else:
            total += v
        return total

    def sum(self, keys):
        """
        Sums the payloads that fall under the dict the keys return. If the keys
         return just a payload, the sum is just that payload.

        WARNING: If += doesn't work on the payload this will fail.

        :param keys: the keys you want to use for accessing the dict
        :return: the sum
        """
        assert len(
            keys) <= self._depth, "To sum values in dict hierarchy, number of keys must be less than or equal to depth of dict"
        return self._sum(self.get(keys))

    def delete(self, keys):
        # TODO document
        # TODO unit test
        assert len(keys) == self._depth, "Number of arguments must equal depth of dict to decrement the payload value"
        result = self._dict
        for key in keys[:-1]:
            result = result[key]
        if keys[-1] in result:
            del result[keys[-1]]

    def __str__(self):
        return str(self._dict)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants.

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.Endpoint import Endpoint
from buttonwood.MarketObjects.Price import PriceFactory
from buttonwood.MarketObjects.Product import Product
import sys

class Market(object):

    def __init__(self, product, endpoint, price_factory, min_qty=1, max_qty=sys.maxint, qty_increment=1):
        """
        Contains all the information needed about a Market.

        A market is where a trade-able asset is traded. In a central limit order book market this would be one order
        book at one venue for one product. This is why the Market is made up of a Product and an Endpoint.

        :param product: Product. the market's product.
        :param endpoint: Endpoint. the market's endpoint.
        :param price_factory: Price.PriceFactory
        :param min_qty: int. the minimum qty that can be traded in the market. Defaults to 1
        :param qty_increment: int. the increment of qty that is allowed in the market. Defaults to 1
        """
        assert isinstance(product, Product)
        assert isinstance(endpoint, Endpoint)
        assert isinstance(price_factory, PriceFactory)
        self._price_factory = price_factory
        self._product = product
        self._endpoint = endpoint
        self._min_qty = min_qty
        self._max_qty = max_qty
        self._qty_increment = qty_increment
        self._hash = hash((self._product, self._endpoint))

    def product(self):
        """
        Get the Product

        :return: Product.
        """
        return self._product

    def endpoint(self):
        """
        Get the Endpoint

        :return:
        """
        return self._endpoint

    def price_factory(self):
        return self._price_factory

    def get_price(self, price_value):
        return self._price_factory.get_price(price_value)

    def min_price_increment(self):
        """
        Gets the minimum price increment of a product. This is the smallest increment that the product trades in.
        :return: Decimal
        """
        return self._price_factory.mpi()

    def min_price_increment_value(self):
        """
        Gets the monetary value associated with the minimum price increment of a product.
        :return: Decimal
        """
        return self._price_factory.mpi_value()

    def mpi(self):
        """
        This is the same as `min_price_increment()'. It is just for the convenience of the developer.

        :return: Decimal
        """
        return self._price_factory.mpi()

    def mpi_value(self):
        """
        This is the same as `min_price_increment_value()'. It is just for the convenience of the developer.

        :return: Decimal
        """
        return self._price_factory.mpi_value()

    def is_valid_price(self, price):
        return self._price_factory.is_valid_price(price)

    def min_qty(self):
        return self._min_qty

    def max_qty(self):
        return self._max_qty

    def qty_increment(self):
        return self._qty_increment

    def is_valid_qty(self, qty):
        return self._min_qty <= qty <= self._max_qty and (qty - self._min_qty) % self._qty_increment == 0

    def __eq__(self, other):
        return isinstance(other, Market) and other.__hash__() == self.__hash__()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%s@%s" % (self.product().name(), self.endpoint().name())

    def to_json(self):
        return {"product": self.product().to_json(), "endpoint": self.endpoint().to_json()}

    def to_detailed_json(self):
        price_info = self._price_factory.to_json()
        return {"product": self.product().to_detailed_json(), "endpoint": self._endpoint.to_json(),
                "price_info": price_info, "min_qty": self._min_qty, "qty_increment": self._qty_increment}

    def __hash__(self):
        return self._hash



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

FAR = 1
FOK = 2
FAK = 3

TIME_IN_FORCE_STRINGS = {FAR: "FAR",
                         FOK: "FOK",
                         FAK: "FAK",
                        }

TIME_IN_FORCE_STR_TO_INT = {v: k for k, v in TIME_IN_FORCE_STRINGS.iteritems()}

def time_in_force_str(int_id):
    assert isinstance(int_id, int)
    if not TIME_IN_FORCE_STRINGS.has_key(int_id):
        raise Exception("%d is an unknown time in force identifier")
    return TIME_IN_FORCE_STRINGS[int_id]

LIMIT = 100
MARKET = 101

ORDER_TYPE_STRINGS = {LIMIT: "Limit",
                      MARKET: "Market",
                     }


"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.MatchSeries import MatchSeries


class MatchSeriesTracker(OrderEventListener):

    def __init__(self, logger):
        OrderEventListener.__init__(self, logger)
        self._aggressor_event_id_to_series = {}
        self._id_to_series = {}
        self._fill_event_id_to_series = {}

    def _handle_fill(self, fill):
        series = self._id_to_series.get(fill.match_id())
        if series is None:
            series = MatchSeries(fill.match_id())
            series.add_fill(fill)
            self._id_to_series[fill.match_id()] = series
            self._aggressor_event_id_to_series[fill.aggressing_command().event_id()] = series
            self._fill_event_id_to_series[fill.event_id()] = series
        else:
            series.add_fill(fill)
            self._fill_event_id_to_series[fill.event_id()] = series

    def match_ids(self):
        return self._id_to_series.keys()

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        self._handle_fill(partial_fill_report)

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        self._handle_fill(full_fill_report)

    def match_series(self, match_id):
        """
        Get the MatchSeries for the given match_id. Can be None if unknown match_id.
        
        :param match_id: unique identifier of MatchSeries 
        :return: Buttonwood.MarketObjects.MatchSeries.MatchSeries
        """
        return self._id_to_series.get(match_id)

    def match_series_for_side(self, side):
        """
        Returns a list of all MatchSeries that had an aggressive side that matches the passed in side
        
        :param side: Buttonwood.MarketObjects.Side.Side
        :return: list() of Buttonwood.MarketObjects.MatchSeries.MatchSeries
        """
        l = []
        for series in self._id_to_series.itervalues():
            if series.aggressor_side() == side:
                l.append(series)
        return l

    def aggressor_match_series(self, aggressor_event_id):
        """
        Return the match series for a given aggressor event id. It will return None if aggressor_event_id is not known.
        
        :param aggressor_event_id: unique identifier of the aggressor event 
        :return: Buttonwood.MarketObjects.MatchSeries.MatchSeries
        """
        return self._aggressor_event_id_to_series.get(aggressor_event_id)

    def fill_event_id_match_series(self, fill_event_id):
        """
        Return the match series for a given fill event id. It will be return None if the fill event id is not known.
        
        :param fill_event_id: unique identifier of the fille vent 
        :return: Buttonwood.MarketObjects.MatchSeries.MatchSeries
        """
        return self._fill_event_id_to_series.get(fill_event_id)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.utils.dicts import NDeepDict


class SubchainTimeAtTopPriorityListener(OrderLevelBookListener):

    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._market_to_subchain_id_to_time = NDeepDict(depth=2, default_value=list)
        self._market_to_side_to_prev_tob_subchain_id = NDeepDict(depth=2, default_value=lambda: None)

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        every time an order book updates, for the side of the causing order chain check the top priority subchain and
         track the amount of time that subchain is at top priority.
         
        Top priority means it is the next to be filled.
        """
        side = causing_order_chain.side()
        top_priority_subchain_id = None
        chains = order_book.order_chains_at_price(side, order_book.best_price(side))
        if len(chains) > 0:
            top_priority_subchain_id = chains[0].most_recent_subchain().subchain_id()
        market = order_book.market()
        use_time = order_book.last_update_time()
        prev_top_priority_subchain_id = self._market_to_side_to_prev_tob_subchain_id.get((market, side))
        if prev_top_priority_subchain_id is not None:
            # if top priority subchain is none and previous top subchain is not None, then close out open time range
            #  of prev
            if top_priority_subchain_id is None:
                l = self._market_to_subchain_id_to_time.get((market, prev_top_priority_subchain_id))
                l[-1] = (l[-1][0], use_time)
            # if top priority id is different than the previous then close out previous and open new one
            elif top_priority_subchain_id != prev_top_priority_subchain_id:
                l_old = self._market_to_subchain_id_to_time.get((market, prev_top_priority_subchain_id))
                l_old[-1] = (l_old[-1][0], use_time)
                l_new = self._market_to_subchain_id_to_time.get((market, top_priority_subchain_id))
                l_new.append((use_time, None))

            # if the same then just keep going, no need to update anything
        else:  # if prev is None then we are just creating a new one
            l = self._market_to_subchain_id_to_time.get((market, top_priority_subchain_id))
            l.append((use_time, None))

        # and set prev to current
        self._market_to_side_to_prev_tob_subchain_id.set((market, side), value=top_priority_subchain_id)

    def time_at_top_priority(self, market, subchain_id, query_time=None):
        """
        Gets the time a subchain spent at top priority. If subchain was never
         at top priority (or didn't even exit) returns 0.

        if query_time is not None (defaults to None) then that time will be used as the ceiling for the query.
         This means if no end time on the last tuple this will be used as the end time. It also means that if any tuple
          overlaps the query time the query time will be used as a cut off.

        :param market: MarketObjects.Market.Market
        :param subchain_id: subchain identifier
        :param query_time: float. The time since epoch of the query (seconds.milli/microseconds)
        :return: float
        """
        total_time = 0
        tob_time_ranges = self._market_to_subchain_id_to_time.get((market, subchain_id))
        if len(tob_time_ranges) == 0:
            return 0
        for tob_time_range in tob_time_ranges[:-1]:
            if query_time is not None and tob_time_range[0] >= query_time:
                return total_time  # no need to go through the rest of list, found the forced end here
            if tob_time_range[1] is None:
                self._logger.warning("%s %s: Cannot correctly calculate top priority time because a non last range end time is None: %s" %
                                     (str(market), str(subchain_id), str(tob_time_ranges)))
                continue

            if query_time is not None and tob_time_range[1] > query_time:
                total_time += query_time - tob_time_range[0]
                return total_time  # no need to go through the rest of list, found the forced end here

            total_time += tob_time_range[1] - tob_time_range[0]

        tob_time_range = tob_time_ranges[-1]
        if query_time is None:
            if tob_time_range[1] is None:
                self._logger.warning(
                    "%s %s: Cannot correctly calculate top priority time because last range end time and query time are None: %s" %
                    (str(market), str(subchain_id), str(tob_time_ranges)))
            else:
                total_time += tob_time_range[1] - tob_time_range[0]
        else:
            if tob_time_range[0] < query_time:
                if query_time >= tob_time_range[1]:
                    total_time += tob_time_range[1] - tob_time_range[0]
                else:
                    total_time += query_time - tob_time_range[0]

        return total_time

    def clean_up_order_chain(self, order_chain):
        market = order_chain.market()
        for subchain in order_chain.subchains():
            self._market_to_subchain_id_to_time.delete([market, subchain.subchain_id()])



class SubchainTimeAtTOBListener(OrderLevelBookListener):
    """
    Tracks how long a subchain is at the top of book.

    This is designed so it can work with mulitiple order books at once.
    """

    # TODO UNIT TEST

    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._market_to_subchain_id_to_time = NDeepDict(depth=2, default_value=list) # this a list of tuples (start_time, end_time)
        self._market_to_side_to_prev_tob_subchain_ids = NDeepDict(depth=2, default_value=set)

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        Every time an orderbook comes in, look at the top of book subchains. Only need to do it for the side that is 
         updating.

        If the same as the previous top of book price then we just update that
         price's last TOB time.

        If not in previous then create a new time (and make sure previous time, if there is one, is valid)
        If in previous then move on with no change.

        If in open list and not in top of book, then close it with the time of update

        """
        # TODO only run the below code if the top of book for the side in question has changed

        side = causing_order_chain.side()

        use_time = order_book.last_update_time()

        if use_time != causing_order_chain.last_update_time():
            self._logger.warn("Order book update time (%.6f) and causing order chain update time (%.6f) do not match!" %
                              (use_time, causing_order_chain.last_update_time()))

        order_chains = order_book.iter_order_chains_at_price(side, order_book.best_price(side))
        market = order_book.market()
        prev_subchain_ids = self._market_to_side_to_prev_tob_subchain_ids.get((market, side))
        found_subchain_ids = set()
        for order_chain in order_chains:
            subchain_id = order_chain.most_recent_subchain().subchain_id()
            found_subchain_ids.add(subchain_id)
            # if in the previous grouping then do nothing
            # if not in the previous grouping then create a new tuple with None for end time
            if subchain_id not in prev_subchain_ids:
                l = self._market_to_subchain_id_to_time.get((market, subchain_id))
                l.append((use_time, None))

        for prev_subchain_id in prev_subchain_ids:
            # if something previously found wasn't found this time around then we need to close it out
            if prev_subchain_id not in found_subchain_ids:
                l = self._market_to_subchain_id_to_time.get((market, prev_subchain_id))
                l[-1] = (l[-1][0], use_time)

        # set previously found to be the new found
        self._market_to_side_to_prev_tob_subchain_ids.set((market, side), value=found_subchain_ids)

    def time_at_top_of_book(self, market, subchain_id, query_time=None):
        """
        Gets the time a subchain spent at top of book. If subchain was never
         at top of book (or didn't even exit) returns 0.

        if query_time is not None (defaults to None) then that time will be used as the ceiling for the query.
         This means if no end time on the last tuple this will be used as the end time. It also means that if any tuple
          overlaps the query time the query time will be used as a cut off.

        :param market: MarketObjects.Market.Market
        :param subchain_id: subchain identifier
        :param query_time: float. The time since epoch of the query (seconds.milli/microseconds)
        :return: float
        """
        total_time = 0
        tob_time_ranges = self._market_to_subchain_id_to_time.get((market, subchain_id))
        if len(tob_time_ranges) == 0:
            return 0
        for tob_time_range in tob_time_ranges[:-1]:
            if query_time is not None and tob_time_range[0] >= query_time:
                return total_time  # no need to go through the rest of list, found the forced end here
            if tob_time_range[1] is None:
                self._logger.warning("%s %s: Cannot correctly calculate TOB time because a non last range end time is None: %s" %
                                     (str(market), str(subchain_id), str(tob_time_ranges)))
                continue

            if query_time is not None and tob_time_range[1] > query_time:
                total_time += query_time - tob_time_range[0]
                return total_time  # no need to go through the rest of list, found the forced end here

            total_time += tob_time_range[1] - tob_time_range[0]

        tob_time_range = tob_time_ranges[-1]
        if query_time is None:
            if tob_time_range[1] is None:
                self._logger.warning(
                    "%s %s: Cannot correctly calculate TOB time because last range end time and query time are None: %s" %
                    (str(market), str(subchain_id), str(tob_time_ranges)))
            else:
                total_time += tob_time_range[1] - tob_time_range[0]
        else:
            if tob_time_range[0] < query_time:
                if query_time >= tob_time_range[1]:
                    total_time += tob_time_range[1] - tob_time_range[0]
                else:
                    total_time += query_time - tob_time_range[0]

        return total_time

    def clean_up_order_chain(self, order_chain):
        market = order_chain.market()
        for subchain in order_chain.subchains():
            self._market_to_subchain_id_to_time.delete([market, subchain.subchain_id()])


"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBooks.OrderLevelBook import OrderLevelBook
from buttonwood.MarketObjects.Side import Side
from buttonwood.MarketObjects.Price import Price

def tob_price_level(order_book, side):
    assert isinstance(order_book, OrderLevelBook)
    assert isinstance(side, Side)
    if order_book is None:
        return None
    return order_book.best_level(side)

def modified_qty_at_price(order_book, side, price, ignore_order_ids, ignore_hidden):
    """
    Returns the qty at the given price and side. Will ignore the size for every order that is in ignore_order_ids.
    
    If ignore hidden is True then hidden qty not included. Otherwise it will be.
    
    If the price does not exist at that side of the order book, will return 0
    
    :param order_book: Buttonwood.MarketObjects.OrderBooks.OrderLevelOrderBook.OrderLevelOrderBook
    :param side: Buttonwood.MarketObjects.Side.Side
    :param price: Buttonwood.MarketObjects.Price.Price
    :param ignore_order_ids: set, list or any other collection that allows for "in" query
    :param ignore_hidden: bool
    :return: 
    """
    # TODO unit test
    assert isinstance(order_book, OrderLevelBook)
    assert isinstance(side, Side)
    assert isinstance(price, Price)
    assert isinstance(ignore_hidden, bool)
    if ignore_order_ids is not None:
        use_ignore_order_ids = ignore_order_ids

        chains = order_book.iter_order_chains_at_price(side, price)
        qty = 0
        for chain in chains:
            if chain.chain_id() not in use_ignore_order_ids:
                qty += chain.visible_qty()
                if not ignore_hidden:
                    qty += chain.hidden_qty()

    else:  # do this else to take advantage of any precalculation that might be done
        qty = order_book.level_at_price(side, price).total_qty()
    return qty






"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json


class Product(object):
    def __init__(self, symbol, name):
        """
        Contains all the information needed about a Product.

        A product + an endpoint == a trade-able asset
        :param symbol: str. the common symbol of the product
        :param name: str. the name of the product (more verbose than product. This is also the unique hash of the product
        """
        assert isinstance(symbol, str)
        assert isinstance(name, str)
        self._symbol = symbol
        self._name = name
        self._identifiers = {}  # used to store different identifiers, like CUSIP, bloomberg universal ID, etc.
        # equality of product ends up getting called pretty frequently so rather than doing a bunch of comparison of a
        #  bunch of getters i'm just going to have one nice tuple and do and do a direct comparison of the tuple
        self._equality_comparitor = (name.lower(), symbol.lower())
        # since this should be an immutable object we can go ahead and create the following
        self.__json = {"symbol": self.symbol(), "name": self.name()}
        self.__str = json.dumps(self.__json)
        self.__hash = hash(name)

    def set_identifier(self, id_type, id_name):
        """
        Add an identifier as key/value pair,
        :param id_type: str. the key of the identifier, such as "cusip"
        :param id_name: str. the value of the identifer, such as "083803"
        :return:
        """
        assert isinstance(id_type, str)
        assert isinstance(id_name, str)
        # keys default to lower for comparison, but not values since case could end up being important
        self._identifiers[id_type.lower()] = id_name

    def get_identifier(self, id_type):
        """
        Gets the name of the identifer assoicated with the type passed in.
        :param id_type: str. the identifier type, such as cusip
        :return: str. the identifier name. can be None if the type is set.
        """
        return self._identifiers.get(id_type.lower())

    def identifiers(self):
        """
        Gets the dictionary of identifiers of the Product.
        :return: dict {str: str}
        """
        return self._identifiers

    def name(self):
        """
        Gets the name of the product. This is a more verbose title of the product when compared to symbol.
        For example:
         *  symbol is 'MSFT'; name is 'microsoft'
         *  symbol is 'GEU5'; name is 'Eurodollar, September 2005'
        :return: str.
        """
        return self._name

    def symbol(self):
        """
        Gets the symbol (aka. abbreviation) commonly used to describe the product. A symbol is less verbose than a name.
         For example:
          *  symbol is 'MSFT'; name is 'microsoft'
          *  symbol is 'GEU5'; name is 'Eurodollar, September 2005'
        :return: str.
        """
        return self._symbol

    def __eq__(self, other):
        identifiers_match = True
        other_identifiers = other.identifiers()
        shared_identifiers = set(self._identifiers).intersection(other_identifiers)
        for shared_identifier in shared_identifiers:
            if self._identifiers[shared_identifier] != other_identifiers[shared_identifier]:
                identifiers_match = False
                break

        return self._equality_comparitor == other._equality_comparitor and identifiers_match

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.__str

    def __hash__(self):
        return self.__hash

    def to_json(self):
        return self.__json

    def to_detailed_json(self):
        # I know that copying the dict is overkill, but it keeps the return json dict from letting someone edit the
        #  identifiers dict accidentally (or on purpose)
        return {"symbol": self.symbol(), "name": self.name(), "identifiers": self._identifiers.copy()}




"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.CancelReasons import CANCEL_TYPES_STRINGS, CANCEL_REASON_STRINGS
from buttonwood.MarketObjects.Events.BasicEvents import BasicEvent
from buttonwood.MarketObjects.Price import Price
from buttonwood.MarketObjects.Market import Market
from buttonwood.MarketObjects.RejectReasons import REJECT_REASON_STRINGS
from buttonwood.MarketObjects.Side import Side
from buttonwood.MarketObjects.Events import OrderEventConstants


class OrderEvent(BasicEvent):
    """
    An OrderEvent is the base class for matching engine order events, which are the messages a customer/user/participant
     sends into the matching engine and the resulting executions the matching engine sends out.

    OrderEvent has two sub categories that are both classes inheriting from it:
      * Commands
      * Execution Reports

    Commands are the events sent into the matching engine. They are the events that express a desire that the matching
     engine then tries to execute on.

    Execution Reports are the events the matching engine sends back after acting on a Command. Every Execution Report
     should have a causing OrderEvent.

    A series of events can be strung together to make up the life cycle of an order. For example:
     * New Order
     * Acknowledgement
     * Cancel Replace
     * Acknowledgement
     * Partial Fill
     * Cancel
     * Cancel Confirm

    This is referred to as an order chain. Events have an order chain ID that
     is used to clearly identify all events that belong to the same order chain.
    """

    def __init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=None):
        """
        The initializer of the base OrderEvent class.

        While it seems like one wouldn't need product, it is great to have here
         for filtering, comparisons, etc.

        `other_key_values` is used for keeping key/value pairs around that don't
         fit the defined arguments. This is an quick and easy way to track meta
         data that a particular matching venue uses/keeps without having to
         create custom versions of the OrderEvent just to keep this data.

        :param event_id: unique identifier of the event
        :param timestamp: float. microsecond time stamp of event. Expecting format of seconds.microseconds (ex: 1234.001123)
        :param chain_id: str or int. the unique identifier fo the orderchain
        :param user_id: str or int. unique identifier of the user who sent the command
        :param market: MarketObjects.Market
        :param other_key_values: dict
        """
        assert isinstance(chain_id, str) or isinstance(chain_id, int)
        assert isinstance(user_id, str) or isinstance(user_id, int)
        assert isinstance(market, Market)
        assert other_key_values is None or isinstance(other_key_values,
                                                      dict), "other_key_values must be none or of type dict"
        BasicEvent.__init__(self, event_id, timestamp)
        self._user_id = user_id
        self._chain_id = chain_id
        self._market = market
        self._other_key_values = {} if other_key_values is None else other_key_values

    def market(self):
        """
        The market of the order event
        :return:
        """
        return self._market

    def user_id(self):
        """
        The user id of the order event

        :return: int or str
        """
        return self._user_id

    def chain_id(self):
        """
        The order chain id of the order event

        :return: int or str
        """
        return self._chain_id

    def get_other_value(self, key):
        """
        Returns the value for the given key from `other_key_values`.

        If the key does not exist in `other_key_values` then returns `None`

        :param key: object
        :return: object. Can be `None`.
        """
        return self._other_key_values.get(key)

    def other_data(self):
        """
        Gets the dictionary of optional other key/value pairs that are stored with the event.
        
        :return: dict. Can be None
        """
        return self._other_key_values

    def _other_values_json(self):
        d = {}
        if self._other_key_values is not None:
            for key, value in self._other_key_values.iteritems():
                if isinstance(value, Market):
                    d[str(key)] = value.to_json()
                elif isinstance(value, Price):
                    d[str(key)] = str(value.price())
                elif hasattr(value, '__dict__'):  # cheap hack to figure out if a primitive or not
                    d[str(key)] = str(value)
                else:
                    d[str(key)] = value
        return d


class OrderCommand(OrderEvent):
    """
    The base class for an order's commands that go into a matching engine. These
     are the events that express a desire
     from a participant that the matching engine attempts to execute.
    """

    def __init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=None):
        OrderEvent.__init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=other_key_values)


class NewOrderCommand(OrderCommand):
    """
    A new order. A new order is necessarily the first event in an order chain.

    If iceberg_peak_qty is None then not taking advantage of iceberg functionality.
    """

    def __init__(self, event_id, timestamp, chain_id, user_id, market, side, time_in_force,
                 price, qty, iceberg_peak_qty=None, limit_or_market=OrderEventConstants.LIMIT, other_key_values=None):
        # TODO documentation
        assert isinstance(side, Side)
        assert isinstance(price, Price)
        assert isinstance(qty, int)
        assert isinstance(limit_or_market, int)
        assert limit_or_market in [OrderEventConstants.MARKET, OrderEventConstants.LIMIT]
        assert iceberg_peak_qty is None or isinstance(iceberg_peak_qty, int)
        assert isinstance(time_in_force, int)
        assert qty > 0, "Qty must be greater than 0"
        assert iceberg_peak_qty is None or iceberg_peak_qty >= 0, "Iceberg Peak Qty must be None or an int >= 0"
        OrderCommand.__init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=other_key_values)
        assert market.is_valid_price(price), "Price %s is not valid for Product %s" % (str(price), str(market))

        self._side = side
        self._price = price
        self._qty = qty
        self._limit_or_market = limit_or_market
        self._iceberg_peak_qty = iceberg_peak_qty
        if self._iceberg_peak_qty is None:
            self._iceberg_peak_qty = qty
        self._time_in_force = time_in_force

    def side(self):
        """
        Get the side.

        :return: MarketObjects.Side
        """
        return self._side

    def price(self):
        """
        Get the price.

        :return: MarketObjects.Price
        """
        return self._price

    def qty(self):
        """
        Get the qty of the new order.

        :return: int
        """
        return self._qty

    def iceberg_peak_qty(self):
        """
        Get the iceberg_peak_qty of the new order

        :return: int
        """
        return self._iceberg_peak_qty

    def time_in_force(self):
        """
        Gets the integer that identifies the order type.

        :return: int
        """
        return self._time_in_force

    def is_far(self):
        """
        Helper function to say if it is a FAR order or not.
        
        :return: bool 
        """
        return self._time_in_force == OrderEventConstants.FAR

    def is_fak(self):
        """
        Helper function to say if it is a FAK order or not.

        :return: bool 
        """
        return self._time_in_force == OrderEventConstants.FAK

    def is_fok(self):
        """
        Helper function to say if it is a FOK order or not.

        :return: bool 
        """
        return self._time_in_force == OrderEventConstants.FOK

    def is_limit_order(self):
        """
        Returns True if the new order is a limit order; false if it is not.

        :return: bool
        """
        return self._limit_or_market == OrderEventConstants.LIMIT

    def is_market_order(self):
        """
        Returns True if the new order is a market order; false if it is not.

        :return: bool
        """
        return self._limit_or_market == OrderEventConstants.MARKET

    def event_type_str(self):
        return "New Order Command"

    def to_json(self):
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'side': int(self.side()),
                                          'price': str(self.price()),
                                          'qty': self.qty(),
                                          'iceberg_peak_qty': self.iceberg_peak_qty(),
                                          'time_in_force': OrderEventConstants.TIME_IN_FORCE_STRINGS[
                                              self.time_in_force()],
                                          'other_key_values': self._other_values_json()}}


class CancelReplaceCommand(OrderCommand):
    def __init__(self, event_id, timestamp, chain_id, user_id, market, side, price, qty, iceberg_peak_qty=None,
                 other_key_values=None):
        # TODO document
        assert isinstance(price, Price)
        assert isinstance(qty, int)
        assert iceberg_peak_qty is None or isinstance(iceberg_peak_qty, int)
        assert iceberg_peak_qty is None or iceberg_peak_qty >= 0, "iceberg_peak_qty cannot be negative."
        assert qty >= 0, "Qty must be greater than 0"
        assert market.is_valid_price(price), "Price %s is not valid for Market %s" % (str(price), str(market))
        OrderCommand.__init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=other_key_values)
        self._side = side
        self._price = price
        self._iceberg_peak_qty = iceberg_peak_qty
        if self._iceberg_peak_qty is None:
            self._iceberg_peak_qty = qty
        self._qty = qty

    def price(self):
        """
        Get the price of the cancel replace.

        :return: MarketObjects.Price
        """
        return self._price

    def side(self):
        """
        Get the side.

        :return: MarketObjects.Side
        """
        return self._side

    def qty(self):
        """
        Get the qty of the cancel replace request.

        :return: int
        """
        return self._qty

    def iceberg_peak_qty(self):
        """
        Get the iceberg_peak_qty of the cancel replace request

        :return: int
        """
        return self._iceberg_peak_qty

    def event_type_str(self):
        return "Cancel Replace Command"

    def to_json(self):
        """
        Get the json dictionary of the cancel replace message

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'price': str(self.price()),
                                          'qty': self.qty(),
                                          'iceberg_peak_qty': self.iceberg_peak_qty(),
                                          'other_key_values': self._other_values_json()}}


class CancelCommand(OrderCommand):
    def __init__(self, event_id, timestamp, chain_id, user_id, market, cancel_type, other_key_values=None):
        # TODO document
        assert isinstance(cancel_type, int)
        OrderCommand.__init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=other_key_values)
        self._cancel_type = cancel_type

    def cancel_type(self):
        """
        Get the cancel type identifier from the Cancel.

        :return: int
        """
        return self._cancel_type

    def cancel_type_str(self):
        """
        Get the cancel type human readable string from the Cancel.

        :return: str
        """
        if self.cancel_type() in CANCEL_TYPES_STRINGS:
            return CANCEL_TYPES_STRINGS[self.cancel_type()]
        return "%d is an unknown cancel reason." % (self.cancel_type())

    def event_type_str(self):
        return "Cancel Command"

    def to_json(self):
        """
        Get the json dictionary of the cancel message

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'cancel_type': self.cancel_type(),
                                          'cancel_type_str': self.cancel_type_str(),
                                          'other_key_values': self._other_values_json()}}


class ExecutionReport(OrderEvent):
    """
    The base class for the execution reports that come out of the matching engine. These are the events that respond
     to the order commands.
    """

    def __init__(self, event_id, timestamp, chain_id, user_id, market, causing_command, other_key_values=None):
        OrderEvent.__init__(self, event_id, timestamp, chain_id, user_id, market, other_key_values=other_key_values)
        self._causing_command = causing_command

    def causing_command(self):
        """
        Get the command that caused the execution report.

        :return: MarketObjects.Events.OrderEvents.OrderCommand
        """
        return self._causing_command


class AcknowledgementReport(ExecutionReport):
    def __init__(self, event_id, timestamp, chain_id, user_id, market, response_to_command, price,
                 qty, iceberg_peak_qty, other_key_values=None):
        """
        An acknowledgement execution report. Has all the standard data for identifying the message, order, etc.

        Also includes the command it is acknowledging (this is for ease of use for listeners to the data as well as to
         systems and usecases that may not be using order chain tracking.

        Price, qty, and iceberg_peak_qty are contained here since what gets ack'd will not always match what was
         requested based on the situation (ex: aggressive orders getting partially filled pre-ack) or the matching
         engine's rules, or the order type (ex: pegged orders starting off at different prices).

        :param event_id: int
        :param timestamp: float
        :param chain_id: int or str
        :param user_id: int or str
        :param market: MarketObjects.Market.Market
        :param response_to_command: MarketObjects.Events.OrderEvents.OrderCommand
        :param price: MarketObjects.Price.Price
        :param qty: int
        :param iceberg_peak_qty: int
        :param other_key_values: dict
        """
        # TODO finish param documentation above
        # TODO finish asserts below
        assert isinstance(price, Price)
        assert isinstance(qty, int)
        assert iceberg_peak_qty is None or isinstance(iceberg_peak_qty, int)
        ExecutionReport.__init__(self, event_id, timestamp, chain_id, user_id, market, response_to_command,
                                 other_key_values=other_key_values)
        self._price = price
        self._qty = qty
        self._iceberg_peak_qty = iceberg_peak_qty
        if self._iceberg_peak_qty is None:
            self._iceberg_peak_qty = qty

    def acknowledged_command(self):
        """
        Get the event that is being acknowledged. Helper function that returns the same as causing_command()

        :return: MarketObjects.Events.OrderEvents.OrderCommand
        """
        return self._causing_command

    def price(self):
        """
        Get the price.

        :return: MarketObjects.Price
        """
        return self._price

    def qty(self):
        """
        Get the qty

        :return: int
        """
        return self._qty

    def iceberg_peak_qty(self):
        """
        Get the iceberg peak qty

        :return: int
        """
        return self._iceberg_peak_qty

    def event_type_str(self):
        return "Acknowledgement Report"

    def to_json(self):
        """
        Get the json dictionary of the acknowledgement

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'response_to_command': self.causing_command().to_json(),
                                          'price': str(self.price()),
                                          'qty': self.qty(),
                                          'iceberg_peak_qty': self.iceberg_peak_qty(),
                                          'other_key_values': self._other_values_json()}}


class RejectReport(ExecutionReport):

    def __init__(self, event_id, timestamp, chain_id, user_id, market, response_to_command, reject_reason,
                 other_key_values=None):
        # TODO document
        assert isinstance(reject_reason, int)
        ExecutionReport.__init__(self, event_id, timestamp, chain_id, user_id, market, response_to_command,
                                 other_key_values=other_key_values)
        self._reject_reason = reject_reason

    def rejected_command(self):
        """
        Get the event that is being rejected. Helper function that returns the same as causing_command()

        :return: MarketObjects.Events.OrderEvents.OrderCommand
        """
        return self._causing_command

    def reject_reason(self):
        """
        Get the reject reason's identifier, which is an int

        :return: int
        """
        return self._reject_reason

    def reject_reason_str(self):
        """
        Get the human readable reject reason string.

        :return: str
        """
        if self.reject_reason() in REJECT_REASON_STRINGS:
            return CANCEL_TYPES_STRINGS[self.reject_reason()]
        return "%d is an unknown cancel reason." % (self.reject_reason())

    def event_type_str(self):
        return "Reject Report"

    def to_json(self):
        """
        Get the json dictionary of the reject report

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'reject_reason': self.reject_reason(),
                                          'reject_reason_str': self.reject_reason_str(),
                                          'response_to_command': self.causing_command().to_json(),
                                          'other_key_values': self._other_values_json()}}


class CancelReport(ExecutionReport):

    def __init__(self, event_id, timestamp, chain_id, user_id, market, cancel_command, cancel_reason,
                 other_key_values=None):
        # TODO document
        assert isinstance(cancel_reason, int)
        ExecutionReport.__init__(self, event_id, timestamp, chain_id, user_id, market, cancel_command,
                                 other_key_values=other_key_values)
        self._cancel_reason = cancel_reason

    def cancel_command(self):
        """
        Get the cancel command that the report is confirming. This is a helper function that is just here for logical
         naming. It simply returns the causing command.
        :return:
        """
        return self.causing_command()

    def cancel_reason(self):
        """
        Get the cancel reason's identifier, which is an int

        :return: int
        """
        return self._cancel_reason

    def cancel_reason_str(self):
        """
        Get the human readable cancel reason string.

        :return: str
        """
        if self.cancel_reason() in CANCEL_REASON_STRINGS:
            return CANCEL_REASON_STRINGS[self.cancel_reason()]
        return "%d is an unknown cancel reason." % (self.cancel_reason())

    def event_type_str(self):
        return "Cancel Report"

    def to_json(self):
        """
        Get the json dictionary of the cancel message

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'cancel_reason': self.cancel_reason(),
                                          'cancel_reason_str': self.cancel_reason_str(),
                                          'cancel_command': "None" if self.causing_command() is None else self.cancel_command().to_json(),
                                          'other_key_values': self._other_values_json()}}


class FillReport(ExecutionReport):

    def __init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command, fill_qty, fill_price,
                 side, match_id, other_key_values=None):
        # TODO document
        assert isinstance(fill_price, Price)
        assert isinstance(fill_qty, int)
        assert isinstance(match_id, int) or isinstance(match_id, str)
        assert isinstance(side, Side)
        ExecutionReport.__init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command,
                                 other_key_values=other_key_values)
        self._fill_price = fill_price
        self._fill_qty = fill_qty
        self._side = side
        self._match_ids = match_id

    def aggressing_command(self):
        """
        Get the aggressing command that the fill was triggered by. This is a helper function that is just here
         for logical naming. It simply returns the causing command.
        :return:
        """
        return self.causing_command()

    def is_aggressor(self):
        """
        Returns True if this fill is for the aggressor that caused the match; returns False otherwise.

        :return: bool
        """
        return self.aggressing_command().chain_id() == self.chain_id()

    def match_id(self):
        """
        Get the match id that the fill is part of.

        :return: int or str.
        """
        return self._match_ids

    def fill_price(self):
        """
        Get the price of the fill.

        :return: MarketObjects.Price
        """
        return self._fill_price

    def fill_qty(self):
        """
        Get the qty filled.

        :return: int
        """
        return self._fill_qty

    def side(self):
        """
        Get the side of the fill.

        :return: MarketObjects.Side
        """
        return self._side

    def event_type_str(self):
        return "Fill Report"

    def to_json(self):
        """
        Get the json dictionary of the fill message.

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'side': int(self.side()),
                                          'fill_price': str(self.fill_price()),
                                          'fill_qty': self.fill_qty(),
                                          'match_id': self.match_id(),
                                          'aggressing_command': self.aggressing_command().to_json(),
                                          'other_key_values': self._other_values_json()}}


class PartialFillReport(FillReport):

    def __init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command, fill_qty, fill_price,
                 side, match_id, leaves_qty, other_key_values=None):
        # TODO document
        assert isinstance(leaves_qty, int)
        FillReport.__init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command, fill_qty,
                            fill_price, side, match_id, other_key_values=other_key_values)
        self._leaves_qty = leaves_qty

    def leaves_qty(self):
        """
        Gets the leaves qty that results from the fill.
        :return:
        """
        return self._leaves_qty

    def event_type_str(self):
        return "Partial Fill Report"

    def to_json(self):
        """
        Get the json dictionary of the partial fill message.

        :return: dict
        """
        return {self.__class__.__name__: {'user_id': self.user_id(),
                                          'chain_id': self.chain_id(),
                                          'event_id': self.event_id(),
                                          'timestamp': self.timestamp(),
                                          'product_name': self.market().product().name(),
                                          'endpoint_name': self.market().endpoint().name(),
                                          'side': int(self.side()),
                                          'fill_price': str(self.fill_price()),
                                          'fill_qty': self.fill_qty(),
                                          'match_id': self.match_id(),
                                          'leaves_qty': self.leaves_qty(),
                                          'aggressing_command': self.aggressing_command().to_json(),
                                          'other_key_values': self._other_values_json()}}


class FullFillReport(FillReport):

    def __init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command, fill_qty, fill_price,
                 side, match_id, other_key_values=None):
        # TODO document
        FillReport.__init__(self, event_id, timestamp, chain_id, user_id, market, aggressing_command, fill_qty,
                            fill_price, side, match_id, other_key_values=other_key_values)

    def event_type_str(self):
        return "Full Fill Report"



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.Events.OrderEvents import NewOrderCommand, CancelReplaceCommand
from buttonwood.MarketObjects.Events.EventChains import OrderEventChain
from buttonwood.MarketObjects.Side import ASK_SIDE
from buttonwood.MarketObjects.Side import BID_SIDE
from buttonwood.utils.dicts import NDeepDict
import operator


class LastTimeTOBListener(OrderLevelBookListener, OrderEventListener):
    """
    Tracks when prices were top of book so that at any given time we can query
     with a price and side and determine when the last time that price was top
     of book.

    If you use the same price but the other side, you get when the last time the
     price would have crossed the book.

    This is designed so it can work with multiple order books at once.
    """
    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        OrderEventListener.__init__(self, logger)
        self._market_to_side_prev_price = NDeepDict(depth=2, default_value=None)
        self._market_side_price_time = NDeepDict(depth=3, default_value=None)  # market to side to price to last time it was top of book
        self._market_side_best_price = NDeepDict(depth=2, default_value=None)
        self._event_id_to_last_time_crossed = dict()
        self._event_id_to_last_time_tob = dict()

    def _update_based_on_new_event(self, market, time):
        for side in [BID_SIDE, ASK_SIDE]:
            # TODO could also loop over every market if we needed to do so here but that impacts performance for what I think is minimal gain
            prev_best_price = self._market_to_side_prev_price.get((market, side))
            if prev_best_price is not None:
                self._market_side_price_time.set((market, side, prev_best_price), value=time)

    def handle_new_order_command(self, new_order_command, resulting_order_chain):
        assert isinstance(new_order_command, NewOrderCommand)
        event_id = new_order_command.event_id()
        market = new_order_command.market()
        time = new_order_command.timestamp()
        self._update_based_on_new_event(market, time)
        price = new_order_command.price()
        side = new_order_command.side()
        if event_id not in self._event_id_to_last_time_crossed:
            self._event_id_to_last_time_crossed[event_id] = self._last_time_crossed(market, side, price)
        if event_id not in self._event_id_to_last_time_tob:
            self._event_id_to_last_time_tob[event_id] = self._last_time_was_tob(market, side, price)

    def handle_cancel_replace_command(self, cancel_replace_command, resulting_order_chain):
        assert isinstance(cancel_replace_command, CancelReplaceCommand)
        event_id = cancel_replace_command.event_id()
        market = cancel_replace_command.market()
        time = cancel_replace_command.timestamp()
        self._update_based_on_new_event(market, time)
        price = cancel_replace_command.price()
        side = cancel_replace_command.side()
        if event_id not in self._event_id_to_last_time_crossed:
            self._event_id_to_last_time_crossed[event_id] = self._last_time_crossed(market, side, price)
        if event_id not in self._event_id_to_last_time_tob:
            self._event_id_to_last_time_tob[event_id] = self._last_time_was_tob(market, side, price)

    def last_time_was_tob(self, event_id):
        last_time_tob = self._event_id_to_last_time_tob[event_id]
        return last_time_tob

    def last_time_crossed(self, event_id):
        last_time_crossed = self._event_id_to_last_time_crossed[event_id]
        if last_time_crossed is not None:
            del self._event_id_to_last_time_crossed[event_id]
        return last_time_crossed

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        """
        Every time an orderbook comes in, look at the top of book price.

        If TOB didn't change at all then just

        If the same as the previous top of book price then we just update that
         price's last TOB time.

        If a new top of book, we need to:
         1) update the previous price to the new time (since it was TOB up to and including this new time)
         2) update the new price to the new time
        """
        time = order_book.last_update_time()
        market = order_book.market()
        # only need to do the side that was updated by the causing order chain as that is only side that changed
        side = causing_order_chain.side()

        best_price = order_book.best_price(side)
        # if current book best price is not none then need to set its time
        if best_price is not None:
            self._market_side_price_time.set((market, side, best_price), value=time)
            prev_best_price = self._market_side_best_price.get((market, side))
            if prev_best_price is None or best_price.better_than(prev_best_price, side):
                self._market_side_best_price.set((market, side), value=best_price)
        # if previous best price is not none and is different than new best price then need to set its time to
        #  update it to include previous price period since it was best price until the change
        prev_best_price = self._market_to_side_prev_price.get((market, side))
        if prev_best_price is not None and prev_best_price != best_price:
            self._market_side_price_time.set((market, side, prev_best_price), value=time)
        # set previous best price to be the current best price
        self._market_to_side_prev_price.set((market, side), value=best_price)

    def _last_time_was_tob(self, market, side, price):
        """
        Returns the last time the price was the top of book
        Will return None if the passed in price has never been top of book for
         market and side during the scope of this listener

        :param market: MarketObjects.Market.Market
        :param side: MarketObjects.Side.Side
        :param price: MarketObjects.Price.Price
        :return: float. (Could be None)
        """
        return self._market_side_price_time.get((market, side, price))

    def _last_time_crossed(self, market, side, price):
        """
        Returns the last time the price for the given side would have crossed the book.
        
        None if it never would have crossed the book or if there was never a book to cross.
        
        :param market: MarketObjects.Market.Market
        :param side: MarketObjects.Side.Side (side of the order we are checking)
        :param price: MarketObjects.Price.Price (price of the order we are checking)
        :return: float. (Could be None)
        """

        resting_side = side.other_side()
        # if book hasn't been established yet, so return None

        # there are a lot of messages that come in that never would have crossed. so just tracking a "best bid and best
        # offer seen all day" would keep us from doing the below sorting and iterating for these and save a ton of time
        best_resting_price = self._market_side_best_price.get((market, resting_side))
        if best_resting_price is None or price.worse_than(best_resting_price, side):
            return None
        else:

            resting_prices_to_time = self._market_side_price_time.get((market, resting_side))

            # sort list based on time, where most recent time (highest number) is first
            sorted_x = sorted(resting_prices_to_time.items(), key=operator.itemgetter(1), reverse=True)
            for x in sorted_x:
                if price.better_or_same_as(x[0], side):
                    return x[1]
        self._logger.error("last_time_crossed has a price that would have crossed today but still has no cross time value. Something is broken!")
        # failsafe return?
        return None

    def clean_up(self, order_chain):
        # if order chain events in map, clean up
        assert isinstance(order_chain, OrderEventChain)
        events = order_chain.events()
        for event in events:
            try:
                del self._event_id_to_last_time_crossed[event.event_id()]
            except:
                pass  # silent fail is okay here. All I'm doing is deleting if it exists in dict. If it doesn't exist no need to delete so failure state is acceptable and faster than checking for existence first.
            try:
                del self._event_id_to_last_time_tob[event.event_id()]
            except:
                pass  # silent fail is okay here. All I'm doing is deleting if it exists in dict. If it doesn't exist no need to delete so failure state is acceptable and faster than checking for existence first.


"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.Events import OrderEventConstants
from buttonwood.MarketObjects.Events.OrderEvents import CancelCommand
from buttonwood.MarketObjects.Events.OrderEvents import CancelReplaceCommand
from buttonwood.MarketObjects.Events.OrderEvents import NewOrderCommand
from buttonwood.utils.dicts import NDeepDict


class OrderEventCountListener(OrderEventListener):

    # TODO unit test
    # TODO document listener

    # COUNT TYPE
    NEW_ORDER = 1
    ACK = 2
    ACK_NEW_ORDERS = 3
    ACK_CANCEL_REPLACE = 4
    CANCEL_REPLACE = 5
    CANCEL_REQUEST = 6
    CANCEL_CONFIRM = 7
    NEW_FAK = 8
    NEW_FAR = 9
    NEW_FOK = 10
    PARTIAL_FILL = 11
    FULL_FILL = 12
    REJECT = 13
    REJECT_NEW = 14
    REJECT_CANCEL_REPLACE = 15
    REJECT_CANCEL = 16
    NEW_LIMIT = 17
    NEW_MARKET = 18
    FAKS_FULLY_FILLED = 19
    FAKS_PARTIALLY_FILLED = 20
    FOKS_FULLY_FILLED = 21
    FARS_FULLY_FILLED_ON_PLACEMENT = 22
    FARS_PARTIALLY_FILLED_ON_PLACEMENT = 23

    def __init__(self, logger):
        OrderEventListener.__init__(self, logger)
        # storing this market -> user -> count type -> count
        self._event_counts = NDeepDict(3, int)

    def get_count(self, market, user_id, count_type):  # get_count(two_year, "user_a", OrderEventCountListener.NEW_FAK)
        return self._event_counts.get([market, user_id, count_type])

    # REQUESTS / COMMANDS IN ######################################

    def handle_new_order_command(self, new_order_command, resulting_order_chain):
        # to be optionally implemented by child class
        self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_ORDER])

        # Time In Force Counts
        if new_order_command.time_in_force() == OrderEventConstants.FAR:
            self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_FAR])
        elif new_order_command.time_in_force() == OrderEventConstants.FAK:
            self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_FAK])
        elif new_order_command.time_in_force() == OrderEventConstants.FOK:
            self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_FOK])

        # Market and limit
        if new_order_command.is_market_order():
            self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_MARKET])
        elif new_order_command.is_limit_order():
            self._event_counts.inc([new_order_command.market(), new_order_command.user_id(), self.NEW_LIMIT])

    def handle_cancel_replace_command(self, cancel_replace_command, resulting_order_chain):
        self._event_counts.inc([cancel_replace_command.market(), cancel_replace_command.user_id(), self.CANCEL_REPLACE])

    def handle_cancel_command(self, cancel_command, resulting_order_chain):
        self._event_counts.inc([cancel_command.market(), cancel_command.user_id(), self.CANCEL_REQUEST])

    # RESPONSES / MESSAGES OUT #####################################

    def handle_acknowledgement_report(self, acknowledgement_report, resulting_order_chain):
        self._event_counts.inc([acknowledgement_report.market(), acknowledgement_report.user_id(), self.ACK])
        if isinstance(acknowledgement_report.acknowledged_command(), NewOrderCommand):
            self._event_counts.inc([acknowledgement_report.market(), acknowledgement_report.user_id(), self.ACK_NEW_ORDERS])
            # if ack comes back for a FAR for a new order, and there is a partial fill in teh orderchain then partially filled on placement
            if resulting_order_chain.time_in_force() == OrderEventConstants.FAR and resulting_order_chain.has_partial_fill():
                self._event_counts.inc([acknowledgement_report.market(), acknowledgement_report.user_id(), self.FARS_PARTIALLY_FILLED_ON_PLACEMENT])
        elif isinstance(acknowledgement_report.acknowledged_command(), CancelReplaceCommand):
            self._event_counts.inc([acknowledgement_report.market(), acknowledgement_report.user_id(), self.ACK_CANCEL_REPLACE])

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        self._event_counts.inc([partial_fill_report.market(), partial_fill_report.user_id(), self.PARTIAL_FILL])

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        self._event_counts.inc([full_fill_report.market(), full_fill_report.user_id(), self.FULL_FILL])

    def handle_cancel_report(self, cancel_report, resulting_order_chain):
        self._event_counts.inc([cancel_report.market(), cancel_report.user_id(), self.CANCEL_CONFIRM])

    def handle_reject_report(self, reject_report, resulting_order_chain):
        self._event_counts.inc([reject_report.market(), reject_report.user_id(), self.REJECT])
        if isinstance(reject_report.rejected_command(), NewOrderCommand):
            self._event_counts.inc([reject_report.market(), reject_report.user_id(), self.REJECT_NEW])
        elif isinstance(reject_report.rejected_command(), CancelReplaceCommand):
            self._event_counts.inc([reject_report.market(), reject_report.user_id(), self.REJECT_CANCEL_REPLACE])
        elif isinstance(reject_report.rejected_command(), CancelCommand):
            self._event_counts.inc([reject_report.market(), reject_report.user_id(), self.REJECT_CANCEL])

    # CLOSE OUT THE CHAIN ##########################################

    def handle_chain_close(self, closed_order_chain):
        if closed_order_chain.has_full_fill():
            if closed_order_chain.time_in_force() == OrderEventConstants.FAK:
                self._event_counts.inc([closed_order_chain.market(), closed_order_chain.user_id(), self.FAKS_FULLY_FILLED])
            elif closed_order_chain.time_in_force() == OrderEventConstants.FOK:
                self._event_counts.inc([closed_order_chain.market(), closed_order_chain.user_id(), self.FOKS_FULLY_FILLED])
            elif closed_order_chain.time_in_force() == OrderEventConstants.FAR:
                # if a FAR has no acknowledgement when fully filled, then it was fully filled on placement
                if not closed_order_chain.has_acknowledgement():
                    self._event_counts.inc([closed_order_chain.market(), closed_order_chain.user_id(), self.FARS_FULLY_FILLED_ON_PLACEMENT])
        elif closed_order_chain.has_partial_fill():
            if closed_order_chain.time_in_force() == OrderEventConstants.FAK:
                self._event_counts.inc([closed_order_chain.market(), closed_order_chain.user_id(), self.FAKS_PARTIALLY_FILLED])



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.OrderBookListeners.OrderLevelBookListener import OrderLevelBookListener
from buttonwood.MarketObjects.Side import BID_SIDE
from buttonwood.MarketObjects.Side import ASK_SIDE


class TopOfBookBeforeEventListener(OrderLevelBookListener):
    """
    Tracks top of book before an event for the market the event occurred in. Keeps track in dict of 
      event_id -> (PriceLevel, PriceLevel), where the first PriceLevel is the bid and second is the ask
        
    """

    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._event_id_to_tob = {}
        self._market_to_previous_tob = {}

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        # set the event to the previous tob
        market = order_book.market()
        tob = self._market_to_previous_tob.get(market)
        event_id = causing_order_chain.most_recent_event().event_id()
        self._event_id_to_tob[event_id] = tob
        # set the previous tob to the new tob (but only do it if TOB changed
        if tob_updated:
            self._market_to_previous_tob[market] = (order_book.best_level(BID_SIDE), order_book.best_level(ASK_SIDE))

    def clean_up_order_chain(self, order_chain):
        for event in order_chain.events():
            del self._event_id_to_tob[event.event_id()]

    def tob_before_event(self, event_id):
        """
        Gets the top of book snapshots as a tuple (bid side PriceLevel, ask side PriceLevel) for the given event_id.
        
        Can be None for either side if the side does not exist in the order book before the event_id.
        
        Can be None instead of a tuple if the event_id is not known.
        
        :param event_id: unique identifier of the event 
        :return: (Buttonwood.MarketObjects.PriceLevel.PriceLevel, Buttonwood.MarketObjects.PriceLevel.PriceLevel)
        """
        return self._event_id_to_tob.get(event_id)


class TopOfBookAfterEventListener(OrderLevelBookListener):
    """
    Tracks top of book after an event for the market the event occurred in. Keeps track in dict of 
      event_id -> (PriceLevel, PriceLevel), where the first PriceLevel is the bid and second is the ask

      if event does not impact the book it doesn't end up getting tracked.
    """

    # TODO UNIT TEST
    def __init__(self, logger):
        OrderLevelBookListener.__init__(self, logger)
        self._event_id_to_tob = {}

    def notify_book_update(self, order_book, causing_order_chain, tob_updated):
        event_id = causing_order_chain.most_recent_event().event_id()
        self._event_id_to_tob[event_id] = (order_book.best_level(BID_SIDE), order_book.best_level(ASK_SIDE))

    def clean_up_order_chain(self, order_chain):
        for event in order_chain.events():
            # not all events cause notify_book_update to be called
            if event.event_id() in self._event_id_to_tob:
                del self._event_id_to_tob[event.event_id()]

    def tob_after_event(self, event_id):
        """
        Gets the top of book snapshots as a tuple (bid side PriceLevel, ask side PriceLevel) for the given event_id.

        Can be None for either side if the side does not exist in the order book after the event_id.

        Can be None instead of a tuple if the event_id is not known.

        :param event_id: unique identifier of the event 
        :return: (Buttonwood.MarketObjects.PriceLevel.PriceLevel, Buttonwood.MarketObjects.PriceLevel.PriceLevel)
        """
        return self._event_id_to_tob.get(event_id)



from distutils.core import setup
from setuptools import find_packages

setup(
  name = 'buttonwood',
  packages = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests",
                                    "*.examples", "*.examples.*", "examples.*", "examples"]),
  version = '1.1.0',
  license='MIT',
  description = 'Buttonwood is a python software package created to help quickly create, (re)build, or analyze markets, market structures, and market participants.',
  author = 'Peter F. Nabicht',
  author_email = 'nabicht@gmail.com',
  url = 'https://github.com/nabicht/Buttonwood',
  download_url = 'https://github.com/nabicht/Buttonwood/archive/v1.1.0.tar.gz',
  keywords = ['markets', 'finance', 'electronic markets', 'microstructure', 'market analysis'],
  install_requires=[
          'nose',
          'cdecimal',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Office/Business :: Financial',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7'
  ],
)


"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from buttonwood.MarketObjects.EventListeners.OrderEventListener import OrderEventListener
from buttonwood.MarketObjects.Events.OrderEvents import OrderCommand
from buttonwood.MarketObjects.Events.OrderEvents import ExecutionReport
from buttonwood.MarketObjects.Events.OrderEvents import NewOrderCommand
from buttonwood.MarketObjects.Events.OrderEvents import CancelReplaceCommand
from buttonwood.MarketObjects.Events.OrderEvents import CancelCommand
from buttonwood.MarketObjects.Events.OrderEvents import AcknowledgementReport
from buttonwood.MarketObjects.Events.OrderEvents import RejectReport
from buttonwood.MarketObjects.Events.OrderEvents import PartialFillReport
from buttonwood.MarketObjects.Events.OrderEvents import FullFillReport
from buttonwood.MarketObjects.Events.OrderEvents import CancelReport
from buttonwood.MarketObjects.Events.EventChains import OrderEventChain
from buttonwood.MarketObjects.OrderBooks.BasicOrderBook import BasicOrderBook
from buttonwood.MarketObjects.OrderBooks.OrderLevelBook import AggregateOrderLevelBook
from buttonwood.MarketObjects.Market import Market
from buttonwood.utils.IDGenerators import MonotonicIntID
from collections import OrderedDict
from collections import defaultdict
import logging


class OrderEventHandler:
    def __init__(self, logger):
        self._event_listeners = OrderedDict()
        self._chain_id_to_chain = {}
        self._market_book_id_to_book = {}
        self._market_to_registered_books = defaultdict(set)
        self._logger = logger
        self._sub_chain_id_generator = MonotonicIntID()

    def register_orderbook(self, market, order_book_id, order_book):
        """
        Allows you to register an order book for a market with a given order book identifier. The identifier allows you
        to name the order book so it is easily retrievable later.

        If the orderbook id already exists for the market an error will be thrown with message "<order book id> is
        already registered for <market name>".  This will prevent overriding an order book that has already been
        registered and populated.

        :param market: MarketObjects.Market.Market
        :param order_book_id: str. The human readable, unique identifier for the orderbook
        :param order_book: MarketObjects.OrderBooks.BasicOrderBook
        """
        assert isinstance(order_book, BasicOrderBook)
        assert isinstance(market, Market)
        assert isinstance(order_book_id, str)

        if isinstance(order_book, AggregateOrderLevelBook):
            for component_order_book in order_book.component_books():
                comp_book_id = "%s Component Book (%s)" % (order_book_id, str(component_order_book.market()))
                self.register_orderbook(component_order_book.market(), comp_book_id, component_order_book)

        if market not in self._market_book_id_to_book:
            self._market_book_id_to_book[market] = {}
        if order_book_id not in self._market_book_id_to_book[market]:
            self._market_book_id_to_book[market][order_book_id] = order_book
        else:
            raise Exception("%s is already registered for %s" % (order_book_id, str(market)))
        self._market_to_registered_books[market].add(order_book)

    def order_book(self, market, order_book_id):
        if market in self._market_book_id_to_book:
            return self._market_book_id_to_book[market].get(order_book_id)
        return None

    def register_event_listener(self, event_listener_id, event_listener):
        """
        Registers and event listener with the given event listener identifier. The event listener identifier allos you
        to name the event listener so it is easily retrievable later.

        If the event listener identifier already exists then an error will be thrown with the message "<event listener
        id> is already registered". This will prevent overriding a listener that has already been
        registered and used.

        The order the event listeners are registered is also the order that they are updated in.

        :param event_listener_id: String. The human readable, unique identifier for the event listener.
        :param event_listener: Buttonwood.MarketObjects.OrderEventListener.OrderEventListener
        """
        assert isinstance(event_listener_id, str)
        assert isinstance(event_listener, OrderEventListener)

        # if the eventlistener that is passed in is an orderbook then WARN the user
        if isinstance(event_listener, BasicOrderBook):
            self._logger.warning(
                "Registering OrderBook as an EventListener! Much downstream logic depends on EventListeners being updated before OrderBooks.")
        if event_listener_id not in self._event_listeners:
            self._event_listeners[event_listener_id] = event_listener
            self._logger.debug("Registered OrderEventListener: '%s': %s" %
                               (event_listener_id, event_listener.__class__.__name__))
        else:
            raise Exception("%s is already registered" % event_listener_id)

    def event_listener(self, event_listener_id):
        """
        Gets the event listener that has been registered wih the passed in id. If the event_listener_id has not been
         registered it will return 100
         
        :param event_listener_id: 
        :return: Buttonwood.MarketObjects.EventListeners.EventListener 
        """
        return self._event_listeners.get(event_listener_id)

    def process(self, event):
        # even though there might not be a handler that is set to debug, the string gets formatted either way,
        #  which is particularly expensive since str(event) actually formats JSON. Thus, the extra level of DEBUG check.
        if logging.DEBUG >= self._logger.getEffectiveLevel():
            self._logger.debug("%s: Processing chain %s event %s: %s" %
                               (self.__class__.__name__, str(event.chain_id()), str(event.event_id()), str(event)))
        order_chain, updated_markets = self._handle_event(event)
        return order_chain, updated_markets

    def _create_new_event_chain(self, new_order_command):
        order_chain = OrderEventChain(new_order_command, self._logger,
                                      subchain_id_generator=self._sub_chain_id_generator)
        return order_chain

    def _handle_new_order_command(self, new_order_command):
        chain_id = new_order_command.chain_id()
        if self._chain_id_to_chain.get(chain_id) is not None:
            chain_id = new_order_command.chain_id()
            self._logger.error("%s: %s for already existing OrderChain %s" %
                               (self.__class__.__name__,
                                new_order_command.event_type_str(),
                                str(self._chain_id_to_chain.get(chain_id))))
        else:
            order_chain = self._create_new_event_chain(new_order_command)
            self._chain_id_to_chain[chain_id] = order_chain
            self._new_order_command_notification(new_order_command, order_chain)

    def _handle_cancel_replace_command(self, cancel_replace_command):
        chain_id = cancel_replace_command.chain_id()
        order_chain = self._chain_id_to_chain.get(chain_id)
        if order_chain is None:
            self._logger.error("%s: Cannot Handle. No OrderChain %s for %s" %
                               (self.__class__.__name__,
                                str(chain_id),
                                cancel_replace_command.event_type_str()))
            return
        order_chain.apply_cancel_replace_command(cancel_replace_command)
        self._cancel_replace_command_notification(cancel_replace_command, order_chain)

    def _handle_cancel_command(self, cancel_command):
        chain_id = cancel_command.chain_id()
        order_chain = self._chain_id_to_chain[chain_id]
        if order_chain is None:
            self._logger.error("%s: Cannot Handle. No OrderChain %s for %s" %
                               (self.__class__.__name__,
                                str(chain_id),
                                cancel_command.event_type_str()))
            return
        order_chain.apply_cancel_command(cancel_command)
        self._cancel_command_notification(cancel_command, order_chain)

    def _handle_acknowledgement_report(self, acknowledgement_report, order_chain):
        order_chain.apply_acknowledgement_report(acknowledgement_report)
        self._acknowledgement_report_notification(acknowledgement_report, order_chain)

    def _handle_reject_report(self, reject_report, order_chain):
        order_chain.apply_reject_report(reject_report)
        self._reject_report_notification(reject_report, order_chain)

    def _handle_cancel_report(self, cancel_report, order_chain):
        order_chain.apply_cancel_report(cancel_report)
        self._cancel_report_notification(cancel_report, order_chain)

    def _handle_partial_fill_report(self, partial_fill_report, order_chain):
        order_chain.apply_partial_fill_report(partial_fill_report)
        self._partial_fill_report_notification(partial_fill_report, order_chain)

    def _handle_full_fill_report(self, full_fill_report, order_chain):
        order_chain.apply_full_fill_report(full_fill_report)
        self._full_fill_report_notification(full_fill_report, order_chain)

    def _apply_to_orderbooks(self, event, order_chain):
        markets_updated = set()
        if order_chain is not None:
            market = event.market()
            if market in self._market_to_registered_books:
                for order_book in self._market_to_registered_books[market]:
                    self._logger.debug("%s: applying %s chain %s event %s to orderbook %s" %
                                       (self.__class__.__name__,
                                        event.__class__.__name__,
                                        str(event.chain_id()),
                                        str(event.event_id()),
                                        order_book.name()))
                    order_book_updated = False
                    if isinstance(event, AcknowledgementReport):
                        order_book_updated, tob_updated = order_book.handle_acknowledgement_report(event, order_chain)
                    elif isinstance(event, CancelReport):
                        order_book_updated, tob_updated = order_book.handle_cancel_report(event, order_chain)
                    elif isinstance(event, PartialFillReport):
                        order_book_updated, tob_updated = order_book.handle_partial_fill_report(event, order_chain)
                    elif isinstance(event, FullFillReport):
                        order_book_updated, tob_updated = order_book.handle_full_fill_report(event, order_chain)
                    else:
                        self._logger.warning("%s: don't know how to handle %s when applying to order book." %
                                             (self.__class__.__name__, event.__class__.__name__))
                    if order_book_updated:
                        markets_updated.add(market)
        return markets_updated

    def _handle_event(self, event):
        """
        A clearing house that takes in an event, figures out what the event is, apply them to the correct order chain,
         and update the correct order book(s) if the event impacts the order book.

        Returns a tuple of the updated data: the order chain and an iterable of markets that had updated order books.

        :param event: Buttonwood.MarketObjects.Events.OrderEvent
        :return: (Buttonwood.MarketObjects.Events.EventChains.OrderEventChain, set of markets)
        """
        # while the command vs execution report check seems unnecessary,
        # it prevents execution reports from having to be checked for each Command
        is_closed_before = False
        is_closed_after = False
        chain_id = event.chain_id()
        markets_with_updated_books = set()
        if isinstance(event, OrderCommand):
            # order is such because in most markets I expect New, than Replace, then cancel in that order
            if isinstance(event, NewOrderCommand):
                # should have been no order_chain if there was a NewOrderCommand, so need to set it
                self._handle_new_order_command(event)
            elif isinstance(event, CancelReplaceCommand):
                self._handle_cancel_replace_command(event)
            elif isinstance(event, CancelCommand):
                self._handle_cancel_command(event)
            else:
                self._logger.error("%s: Cannot handle unknown OrderCommand: %s" %
                                   (self.__class__.__name__,
                                    event.__class__.__name__))
        elif isinstance(event, ExecutionReport):
            order_chain = self._chain_id_to_chain.get(chain_id)
            if order_chain is None:
                self._logger.warning("%s: Cannot Handle. No OrderChain %s for %s" %
                                     (self.__class__.__name__,
                                      str(chain_id),
                                      event.event_type_str()))
                return
            # closes can only happen as result of an ExecutionReport
            is_closed_before = not order_chain.is_open()
            # the ordering is because I expect to see more Acks, cancel confirms, then partial fills, then full fills, then rejects in most markets
            if isinstance(event, AcknowledgementReport):
                self._handle_acknowledgement_report(event, order_chain)
            elif isinstance(event, CancelReport):
                self._handle_cancel_report(event, order_chain)
            elif isinstance(event, PartialFillReport):
                self._handle_partial_fill_report(event, order_chain)
            elif isinstance(event, FullFillReport):
                self._handle_full_fill_report(event, order_chain)
            elif isinstance(event, RejectReport):
                self._handle_reject_report(event, order_chain)
            else:
                self._logger.error("%s: Cannot handle unknown ExecutionReport: %s" %
                                   (self.__class__.__name__,
                                    event.__class__.__name__))
                return None, markets_with_updated_books
            is_closed_after = not order_chain.is_open()

            # apply to the registered order book (if one is registered)
            # only execution reports impact an order book so only applying to order book here
            markets_with_updated_books = self._apply_to_orderbooks(event, order_chain)
        else:
            self._logger.error("%s: Cannot handle unknown Event: %s" %
                               (self.__class__.__name__,
                                event.__class__.__name__))
            return None, markets_with_updated_books

        # order_chain close notifications
        order_chain = self._chain_id_to_chain.get(chain_id)
        if order_chain is not None and not is_closed_before and is_closed_after:
            self._close_chain_notification(order_chain)
            # if closed then no longer need to keep it in map:
            del self._chain_id_to_chain[chain_id]
        return order_chain, markets_with_updated_books

    def _new_order_command_notification(self, new_order_command, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_new_order_command(new_order_command, resulting_order_chain)

    def _cancel_replace_command_notification(self, cancel_replace_command, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_cancel_replace_command(cancel_replace_command, resulting_order_chain)

    def _cancel_command_notification(self, cancel_command, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_cancel_command(cancel_command, resulting_order_chain)

    def _acknowledgement_report_notification(self, acknowledgement_report, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_acknowledgement_report(acknowledgement_report, resulting_order_chain)

    def _partial_fill_report_notification(self, partial_fill_report, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_partial_fill_report(partial_fill_report, resulting_order_chain)

    def _full_fill_report_notification(self, full_fill_report, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_full_fill_report(full_fill_report, resulting_order_chain)

    def _cancel_report_notification(self, cancel_report, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.handle_cancel_report(cancel_report, resulting_order_chain)

    def _reject_report_notification(self, reject_report, resulting_order_chain):
        for listener in self._event_listeners.itervalues():
            listener.notify_reject_reort(reject_report, resulting_order_chain)

    def _close_chain_notification(self, closed_order_chain):
        if closed_order_chain.is_open():
            self._logger.error("%s: _close_chain_notification called for order chain %s and it is NOT closed! Not notifying listeners." %
                               (self.__class__.__name__, str(closed_order_chain.chain_id())))
            return
        for listener in self._event_listeners.itervalues():
            listener.handle_chain_close(closed_order_chain)

    def chain_ids(self):
        return self._chain_id_to_chain.keys()

    def order_chain(self, chain_id):
        return self._chain_id_to_chain.get(chain_id)



"""
This file is part of Buttonwood.

Buttonwood is a python software package created to help quickly create, (re)build, or 
analyze markets, market structures, and market participants. 

MIT License

Copyright (c) 2016-2019 Peter F. Nabicht

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class OrderEventListener(object):

    def __init__(self, logger):
        self._logger = logger

    # REQUESTS / COMMANDS IN ######################################

    def handle_new_order_command(self, new_order_command, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a new order
        command being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        NewOrderCommand has been applied.

        :param new_order_command: MarketObjects.Events.OrderEvents.NewOrderCommand
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_cancel_replace_command(self, cancel_replace_command, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a cancel
        replace command being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        CancelReplaceCommand has been applied.

        :param cancel_replace_command: MarketObjects.Events.OrderEvents.CancelReplaceCommand
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_cancel_command(self, cancel_command, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a cancel
        command being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        CancelCommand has been applied.

        :param cancel_command: MarketObjects.Events.OrderEvents.CancelCommand
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    # RESPONSES / MESSAGES OUT #####################################

    def handle_acknowledgement_report(self, acknowledgement_report, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a acknowledgement
        execution report being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        AcknowledgementReport has been applied.

        :param acknowledgement_report: MarketObjects.Events.OrderEvents.AcknowledgementReport
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_partial_fill_report(self, partial_fill_report, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a partial fill
        execution report being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        PartialFillReport has been applied.

        :param partial_fill_report: MarketObjects.Events.OrderEvents.PartialFillReport
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_full_fill_report(self, full_fill_report, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a full fill
        execution report being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        FullFillReport has been applied.

        :param full_fill_report: MarketObjects.Events.OrderEvents.FullFillReport
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_cancel_report(self, cancel_report, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a cancel
        execution report being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        CancelReport has been applied.

        :param cancel_report: MarketObjects.Events.OrderEvents.CancelReport
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def handle_reject_report(self, reject_report, resulting_order_chain):
        """
        The OrderEventListener's call back for being notified of a full fill
        execution report being added to the order chain.

        The resulting_order_chain is the OrderChain object after the
        FullFillReport has been applied.

        :param reject_report: MarketObjects.Events.OrderEvents.RejectReport
        :param resulting_order_chain: MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    # CLOSE OUT THE CHAIN ##########################################

    def handle_chain_close(self, closed_order_chain):
        """
        The OrderEventListener's call back for being notified of a closed order
        chain.

        When the order chain is closed there might be special logic in the
        analytic that needs to run, or at the very least some cleanup that could
        occur.

        :param closed_order_chain: Buttonwood.MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass

    def clean_up(self, order_chain):
        """
        This is an optional function to be used to clean up memory / data
         structures / etc. when an order_chain is done with.

        The idea is that when a user of the metrics is done with the data for a
         given order_chain, it can call "clean_up" on the listeners and data that
         is no longer needed can be deleted.

        If this function is called before all uses of the OrderEventListener
         are done with the order chain then those uses won't be able to get the
         needed data.

        If a listener tracks data by subchain_id then the cleanup will need to
         go through all subchains in the order chain and do cleanup for each one.

        If an inheritng class of OrderEventListener does not implement this
         function then simply nothing is done to do the clean up.
        
        :param order_chain: Buttonwood.MarketObjects.Events.EventChains.OrderEventChain
        """
        # to be optionally implemented by child class
        pass
